"""
authlib
~~~~~~~

The ultimate Python library in building OAuth 1.0, OAuth 2.0 and OpenID
Connect clients and providers. It covers from low level specification
implementation to high level framework integrations.

:copyright: (c) 2017 by Hsiaoming Yang.
:license: BSD, see LICENSE for more details.
"""

from .consts import author
from .consts import homepage
from .consts import version

__version__ = version
__homepage__ = homepage
__author__ = author
__license__ = "BSD-3-Clause"
